QStackedWidget, QWidget {
    background-color: #EEE;
    color: black;
}

QToolButton:hover {
    background-color: #cccccc;
    border: #999999;
}

QTextEdit {
    background-color: white;
    color: black;
}

QToolBar, QToolButton {
    background: transparent;
}

QListWidget {
    background-color: white;
    color: black;
}
