QStackedWidget, QWidget, QApplication {
    background-color: #222;
    color: #CCC;
}

QToolButton:hover {
    background-color: #333;
    border: #CCC;
}

QTextEdit {
    background-color: #222;
    color: #CCC;
}

QToolBar, QToolButton {
    background: transparent;
}

QListWidget {
    background-color: #222;
    color: #CCC;
}
